/*
 * Resource server functionality.
 *
 * Copyright (C) 2018-2022, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/types.h>

#include <pthread-l4.h>
#include <pthread.h>

#include "resource_server.h"



/* Convenience server methods. */

/* Bind to a named IPC gate capability. */

long ResourceServer::bind(const char *name)
{
  _config->server = l4re_env_get_cap(name);
  return L4_EOK;
}

/* Start in the same thread indicating whether the server can be finalised. If
   so, deletion notifications will be used. */

long ResourceServer::start(bool finalisation)
{
  resource_init_config(_config, _resource);
  _config->thread = pthread_l4_cap(pthread_self());

  /* NOTE: On MIPS32, at least, in a payload started by libexec, the main thread
           is not necessarily returned correctly. */

  if (l4_is_invalid_cap(_config->thread))
    _config->thread = l4re_env()->main_thread;

  if (finalisation)
  {
    _config->finaliser = resource_same_thread_finaliser;
    _config->notifications = 1;
    _config->auto_deletion = 1;
  }

  return resource_start_config(_config, _resource);
}

/* Start serving a resource in an existing thread. The resource will not be
   finalised. */

long ResourceServer::start_in_thread(l4_cap_idx_t thread)
{
  resource_init_config(_config, _resource);
  resource_set_config_threaded(_config, thread, 1, 0, 0);

  return resource_start_config(_config, _resource);
}

/* Start serving a resource in a new thread. */

long ResourceServer::start_thread(bool finalisation, bool auto_deletion)
{
  pthread_t thread;
  pthread_attr_t attr;
  long err;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  resource_init_config(_config, _resource);

  err = pthread_create(&thread, &attr, ipc_server_start_mainloop, _config);
  if (err)
    return err;

  resource_set_config_threaded(_config, pthread_l4_cap(thread), 1, finalisation,
                               auto_deletion);

  return resource_start_config(_config, _resource);
}

/* A convenience method starting a thread and returning the server capability
   employed via the given parameter. */

long ResourceServer::start_thread(l4_cap_idx_t *server, bool finalisation,
                                  bool auto_deletion)
{
  long err = start_thread(finalisation, auto_deletion);

  if (!err)
    *server = _config->server;

  return err;
}



/* Initialise a server configuration for a resource. */

void resource_init_config(ipc_server_config_type *config, Resource *resource)
{
  ipc_server_default_config_type default_config = resource->config();

  config->handler_obj = resource->interface();
  config->finaliser_obj = resource;
  config->expected_items = default_config.expected_items;
  config->handler = default_config.handler;
}

/* Set a configuration to be threaded. */

void resource_set_config_threaded(ipc_server_config_type *config,
                                  l4_cap_idx_t thread, int separate_thread,
                                  int finalisation, int auto_deletion)
{
  /* Only invoke the finaliser where auto-deletion is employed. This assumes
     that a single configuration is in use and thus a single resource that can
     be managed in the server finalisation process. Where multiple
     configurations are involved with an IPC gate, server finalisation does not
     attempt to finalise all these configurations and associated resources, and
     it therefore makes more sense to perform such finalisation elsewhere. */

  if (auto_deletion)
    config->finaliser = resource_thread_finaliser;

  config->config_thread = separate_thread;
  config->thread = thread;
  config->notifications = finalisation;
  config->auto_deletion = auto_deletion;
}

/* Activate a resource and start a server for it. */

long resource_start_config(ipc_server_config_type *config, Resource *resource)
{
  resource->activate();
  long err = ipc_server_start_config(config);

  /* Discard any server resources if starting it failed. */

  if (err)
  {
    ipc_server_finalise_config(config);
    ipc_server_discard_thread(config);
  }

  return err;
}



/* A finaliser for exposed resources in the same thread. */

void resource_same_thread_finaliser(ipc_server_config_type *config)
{
  Resource *resource = reinterpret_cast<Resource *>(config->finaliser_obj);

  /* Close but do not delete the resource since it is assumed that it is being
     managed by the thread. */

  resource->close();

  /* Release the capabilities. */

  ipc_server_finalise_config(config);
  delete config;
}

/* A finaliser for exposed resources in a different thread. */

void resource_thread_finaliser(ipc_server_config_type *config)
{
  Resource *resource = reinterpret_cast<Resource *>(config->finaliser_obj);

  /* Close and delete the resource. */

  resource->close();
  delete resource;

  /* Release the capabilities. */

  ipc_server_finalise_config(config);
  delete config;
}

// vim: tabstop=2 expandtab shiftwidth=2
